/* This file is part of the Project Athena Zephyr Notification System.
 * It contains source for the ZInitialize function.
 *
 *	Created by:	Robert French
 *
 *	$Id: ZInit.c,v 1.27 1999/01/22 23:19:14 ghudson Exp $
 *
 *	Copyright (c) 1987, 1991 by the Massachusetts Institute of Technology.
 *	For copying and distribution information, see the file
 *	"mit-copyright.h". 
 */

#ifndef lint
static char rcsid_ZInitialize_c[] =
    "$Zephyr: /afs/athena.mit.edu/astaff/project/zephyr/src/lib/RCS/ZInitialize.c,v 1.17 89/05/30 18:11:25 jtkohl Exp $";
#endif

#include <internal.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <config.h>
#ifdef HAVE_SYS_SOCKIO_H
#include <sys/sockio.h>
#endif
#ifdef HAVE_IFADDRS_H
#include <ifaddrs.h>
#endif
#ifdef HAVE_NET_IF_H
#include <net/if.h>
#endif
#ifdef HAVE_KRB4
#include <krb_err.h>
#endif

#ifdef HAVE_SA_LEN
#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#endif
#define ifreq_size(i) max(sizeof(struct ifreq),\
     sizeof((i).ifr_name)+(i).ifr_addr.sa_len)
#else
#define ifreq_size(i) sizeof(struct ifreq)
#endif /* HAVE_SA_LEN*/

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif

Code_t ZInitialize()
{
    struct servent *hmserv;
    struct hostent *hostent;
    char hostname[MAXHOSTNAMELEN];
    struct in_addr servaddr;
    struct sockaddr_in sin;
    int s, sinsize = sizeof(sin);
    Code_t code;
    ZNotice_t notice;
#ifdef HAVE_KRB4
    char *krealm = NULL;
    int krbval;
    char d1[ANAME_SZ], d2[INST_SZ];

    initialize_krb_error_table();
#endif

    initialize_zeph_error_table();
    
    (void) memset((char *)&__HM_addr, 0, sizeof(__HM_addr));

    __HM_addr.sin_family = AF_INET;
#ifdef HAVE_SA_LEN
    __HM_addr.sin_len = sizeof (struct sockaddr_in);
#endif
    __HM_addr.sin_addr.s_addr = htonl (0x7f000001L);

    hmserv = (struct servent *)getservbyname(HM_SVCNAME, "udp");
    __HM_addr.sin_port = (hmserv) ? hmserv->s_port : HM_SVC_FALLBACK;

    __HM_set = 0;

    /* Initialize the input queue */
    __Q_Tail = NULL;
    __Q_Head = NULL;
    
    /* if the application is a server, there might not be a zhm.  The
       code will fall back to something which might not be "right",
       but this is is ok, since none of the servers call krb_rd_req. */

    servaddr.s_addr = INADDR_NONE;
    if (! __Zephyr_server) {
       if ((code = ZOpenPort(NULL)) != ZERR_NONE)
	  return(code);

       if ((code = ZhmStat(NULL, &notice)) != ZERR_NONE)
	  return(code);

       ZClosePort();

       /* the first field, which is NUL-terminated, is the server name.
	  If this code ever support a multiplexing zhm, this will have to
	  be made smarter, and probably per-message */

#ifdef HAVE_KRB4
       krealm = krb_realmofhost(notice.z_message);
#endif
       hostent = gethostbyname(notice.z_message);
       if (hostent && hostent->h_addrtype == AF_INET)
	   servaddr = *(struct in_addr *)hostent->h_addr;

       ZFreeNotice(&notice);
    }

#ifdef HAVE_KRB4
    if (krealm) {
	strcpy(__Zephyr_realm, krealm);
    } else if ((krb_get_tf_fullname(TKT_FILE, d1, d2, __Zephyr_realm)
		!= KSUCCESS) &&
	       ((krbval = krb_get_lrealm(__Zephyr_realm, 1)) != KSUCCESS)) {
	return (krbval);
    }
#else
    strcpy(__Zephyr_realm, "local-realm");
#endif

    __My_addr.s_addr = INADDR_NONE;
    /* Should we query Zephyr variables for an override?  */
    if (servaddr.s_addr != INADDR_NONE) {
	/* Try to get the local interface address by connecting a UDP
	 * socket to the server address and getting the local address.
	 * Some broken operating systems (e.g. Solaris 2.0-2.5) yield
	 * INADDR_ANY (zero), so we have to check for that. */
	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s != -1) {
	    memset(&sin, 0, sizeof(sin));
	    sin.sin_family = AF_INET;
	    sin.sin_addr = servaddr;
	    sin.sin_port = HM_SRV_SVC_FALLBACK;
	    if (connect(s, (struct sockaddr *) &sin, sizeof(sin)) == 0
		&& getsockname(s, (struct sockaddr *) &sin, &sinsize) == 0
		&& sin.sin_addr.s_addr != 0)
		__My_addr = sin.sin_addr;
	    close(s);
	}
    }
    if (__My_addr.s_addr == INADDR_NONE) {
	/* If the structure of this part looks a little wacky, that's
	   because it's a stripped-down version of the current krb5
	   code, which was a bit more complicated than Zephyr needs,
	   because it makes a couple of passes, handles multiple
	   address families, needs to record all the addresses, etc.

	   In particular, the second branch could avoid growing its
	   buffer whennot on Solaris if the data already acquired
	   includes an appropriate IPv4 address.

	   This version is less preferable anyways, compared to the
	   above, because it has no idea which interface on a
	   multihomed system would be used to reach the server.  The
	   connected-socket approach has at least half a chance of
	   getting something useful based on the routing table.  */
#ifdef HAVE_IFADDRS_H
	struct ifaddrs *ifp_head, *ifp;
	int match;

	if (getifaddrs (&ifp_head) < 0)
	    return errno;
	for (ifp = ifp_head; ifp; ifp = ifp->ifa_next) {
	    if ((ifp->ifa_flags & IFF_UP) == 0)
		continue;
	    if (ifp->ifa_flags & IFF_LOOPBACK)
		continue;
	    if (ifp->ifa_addr->sa_family != AF_INET)
		continue;
	    /* Okay, looks good.  */
	    __My_addr = ((struct sockaddr_in *)(ifp->ifa_addr))->sin_addr;
	    break;
	}
	freeifaddrs (ifp_head);
#else
	struct ifreq *ifr, ifreq;
	int s, code;
	int est_if_count = 8;
	size_t est_ifreq_size;
	char *buf = 0;
	size_t current_buf_size = 0, size, n, i;
	struct ifconf ifc;
#ifdef SIOCGSIZIFCONF
	int ifconfsize = -1;
#endif
#ifdef SIOCGIFNUM
	int numifs = -1;
#endif

	s = socket (AF_INET, SOCK_DGRAM, 0);
	if (s < 0)
	    goto no_socket;

	/* At least on NetBSD, an ifreq can hold an IPv4 address, but
	   isn't big enough for an IPv6 or ethernet address.  So add a
	   little more space.  */
	est_ifreq_size = sizeof (struct ifreq) + 8;
#ifdef SIOCGSIZIFCONF
	code = ioctl (s, SIOCGSIZIFCONF, &ifconfsize);
	if (!code) {
	    current_buf_size = ifconfsize;
	    est_if_count = ifconfsize / est_ifreq_size;
	}
#elif defined (SIOCGIFNUM)
	code = ioctl (s, SIOCGIFNUM, &numifs);
	if (!code && numifs > 0)
	    est_if_count = numifs;
#endif
	if (current_buf_size == 0)
	    current_buf_size = est_ifreq_size * est_if_count;
	buf = malloc (current_buf_size);
	if (buf == NULL)
	    return errno;

    ask_again:
	size = current_buf_size;
	ifc.ifc_len = size;
	ifc.ifc_buf = buf;
	memset (buf, 0, size);
	code = ioctl (s, SIOCGIFCONF, (char *) &ifc);
	size = ifc.ifc_len;
	if (code < 0) {
	    close (s);
	    free (buf);
	    goto punt;
	}
	/* Test that the buffer was big enough that another ifreq could've
	   fit easily, if the OS wanted to provide one.  That seems to be
	   the only indication we get, complicated by the fact that the
	   associated address may make the required storage a little
	   bigger than the size of an ifreq.  */
	if (current_buf_size - size < sizeof (struct ifreq) + 40
#ifdef SIOCGSIZIFCONF
	    && ifconfsize <= 0
#elif defined (SIOCGIFNUM)
	    && numifs <= 0
#endif
	    ) {
	    size_t new_size;
	    void *newptr;

	    est_if_count *= 2;
	    new_size = est_ifreq_size * est_if_count;
	    newptr = realloc (buf, new_size);
	    if (newptr == 0 && new_size != 0) {
		free (buf);
		goto punt;
	    }
	    buf = newptr;
	    if (buf == 0)
		return errno;
	    current_buf_size = new_size;
	    goto ask_again;
	}

	n = size;

	for (i = 0; i < n; i+= ifreq_size(*ifr) ) {
	    ifr = (struct ifreq *)((caddr_t) buf+i);
	    
	    strncpy(ifreq.ifr_name, ifr->ifr_name, sizeof (ifreq.ifr_name));
	    if (ioctl (s, SIOCGIFFLAGS, (char *)&ifreq) < 0)
		continue;

#ifdef IFF_LOOPBACK
	    /* None of the current callers want loopback addresses.  */
	    if (ifreq.ifr_flags & IFF_LOOPBACK)
		continue;
#endif
	    /* Ignore interfaces that are down.  */
	    if ((ifreq.ifr_flags & IFF_UP) == 0)
		continue;

	    if (((struct sockaddr *)(&ifr->ifr_addr))->sa_family != AF_INET)
		continue;

	    __My_addr = ((struct sockaddr_in *)(&ifr->ifr_addr))->sin_addr;
	    break;
	}

    punt:
	close (s);
    no_socket:
	free (buf);
#endif /* not HAVE_IFADDRS_H */
    }
    if (__My_addr.s_addr == INADDR_NONE) {
	/* We couldn't figure out the local interface address by the
	 * above method.  Try by resolving the local hostname.  (This
	 * is a pretty broken thing to do, and unfortunately what we
	 * always do on server machines.) */
	if (gethostname(hostname, sizeof(hostname)) == 0) {
	    hostent = gethostbyname(hostname);
	    if (hostent && hostent->h_addrtype == AF_INET)
		__My_addr = *(struct in_addr *)hostent->h_addr;
	}
    }
    /* If the above methods failed, zero out __My_addr so things will
     * sort of kind of work. */
    if (__My_addr.s_addr == INADDR_NONE)
	__My_addr.s_addr = 0;

    /* Get the sender so we can cache it */
    (void) ZGetSender();

    return (ZERR_NONE);
}

